#include "common.h"

/* Common code (root and node) */

#define SIZE    80
#define DEBUG           1


int array[SIZE];      /* Array to sort in root processor */

/*{{{  initialize*/
/* Sets randomly the values for the array */

void initialize(int *v) {
  unsigned i;

   for(i = 0; i < SIZE; i++) {
     RAND(v[i]);
   #ifdef DEBUG
     PRINT3("\n%d %d",i,v[i]);
   #endif
   }
}

/*}}}  */

/*{{{  input/output*/

/****************************************************************************/
void press_key(void) {
  char tecla;

  GPRINT1("\nPULSA <<RETURN>> PARA CONTINUAR");
  tecla = GETC(stdin);
  GPRINT1("\n\n");
}

/****************************************************************************/
/* Escribe el array */
void write_array(int *v) {
  unsigned i;

  for(i = 0; i < SIZE; i++)
    PRINT3("v[%u] = %i\n", i, v[i]);
}


/*}}}  */

/*{{{  test if errors*/
void test(int *v) {
  int i;

  for (i = 0; i < SIZE - 1; i++)
    if (v[i] > v[i + 1]) {
      PRINT5("\nERROR: Array not sorted. Elements %d and %d. (%d, %d)\n",
	    i, i+1, v[i], v[i+1]);
      return;
    }
  GPRINT1("Array sorted.\n");
}
/*}}}  */

/*{{{  quicksortseq(int *v, int first, int last)*/
/* Quicksort clsico (Wirth) */
void quicksortseq(int *v, int first, int last) {
	int i, j;
	int pivot;
	int temp;

	i = first;
	j = last;
	pivot = v[(first + last) / 2];
	do {
		while (v[i] < pivot)
			i++;
		while (pivot < v[j])
			j--;
		if (i <= j) {
			temp = v[i];    /* intercambiar v[i] y v[j] */
			v[i] = v[j];
			v[j] = temp;
			i++;
			j--;
		}
	} while (i <= j);
	if (first < j)
		quicksortseq(v, first, j);
	if (i < last)
		quicksortseq(v, i, last);
}
/*}}}  */

/*{{{  partition(int *v, int *i, int *j, int first, int last)*/

/****************************************************************************/
/* Partition clsico (el de Hoare) */
/****************************************************************************/
void partition(int *v, int *i, int *j, int first, int last) {
	register int pivot, temp;

	*i = first;
	*j = last;
	pivot = v[(first + last) / 2];
	while (*i <= *j) {
		while (v[*i] < pivot)
			(*i)++;
		while (pivot < v[*j])
			(*j)--;
		if (*i <= *j) {
			temp = v[*i];   /* intercambiar v[i] y v[j] */
			v[*i] = v[*j];
			v[*j] = temp;
			(*i)++;
			(*j)--;
		}
	}
}
/*}}}  */

/*{{{  find*/
/* Partition equilibrado (Brinch Hansen).
	Basndose en partition divide el array en dos mitades de modo que ambas
	tengan igual tamao. De este modo, la divisin de carga entre los
	procesadores ser equitativa.
	BIBLIOGRAFIA:   "Parallel Divide and Conquer" Per Brinch Hansen
						School of Computer and Information Science
						Syracuse University, Syracuse, New York, 13244, U.S.A.
						December 1991
*/
void find(int *v, int first, int last, int middle) {
	int left, right, i, j;

	left = first;
	right = last;
	while (left < right) {
		partition(v, &i, &j, left, right);
		if (middle <= j)
			right = j;
		else
			if (i <= middle)
				left = i;
			else
				left = right;
	}
}
/*}}}  */

/*{{{  parallel quicksort*/
void qs(int *v, int first, int last) {
  int middle;
  int size;

  if (num_processors > 1) {
    if (first < last) {
      middle = (first + last) / 2;
      size = (middle - first + 1) * sizeof(int);
      find(v, first, last, middle);

      /*
      PARREDUCE(qs(v, first, middle),
		qs(v, middle + 1, last),
		(v + middle + 1), (size)));
      */

      /*
      PAR(qs(v, first, middle),
	  (v + first), (size),
	  qs(v, middle + 1, last),
	  (v + middle + 1), (size)));
      */

      /*{{{  parcalls*/
      num_processors /= 2;
      bit--;
      if ((*name & (1 << bit)) == 0) {
	qs(v, first, middle);
	ChanIn(d_in[bit], (v + middle + 1), (size));
	ChanOut(d_out[bit], (v + first), (size));
      }
      else {
	qs(v, middle + 1, last);
	ChanOut(d_out[bit], (v + middle + 1), (size));
	ChanIn(d_in[bit], (v + first), (size));
      }
      num_processors *= 2;
      bit++;
      }
      /*}}}  */
    }
    else {
      quicksortseq(v, first, last);
    }
}
/*}}}  */

int main(void) {

  /*{{{  variables*/
  int i;
  clock_t ini_time, fin_time, dif_time;           /* Time measurement */
  int first, last;                                /* Range to sort in the array */
  /*}}}  */

  INITIALIZE;


  ini_time = clock();

  /*{{{  debug*/
  #ifdef DEBUG
     PRINT4("dim = %d Processor %d is root... Level: %d\n",dim, *name, level);
     for (i = 1; i < num_processors; i++)
       PRINT3("Processor %d level: %d\n", i, node_level(i));
  #endif
  /*}}}  */

  initialize(array);

  /*{{{  debug*/
  #ifdef DEBUG
     GPRINT1("Array generated in root...\n");
     write_array(array);
     press_key();
  #endif
  /*}}}  */

  first = 0;
  last = SIZE - 1;

  qs(array, first, last);

  /*{{{  time*/
  fin_time = clock();
  dif_time = fin_time - ini_time;
  PRINT3("\nElapsed time: %ld. (%lf)\n", dif_time,
  (double) dif_time / CLOCKS_PER_SEC);
  /*}}}  */

  /*{{{  test array and terminate*/
  test(array);
  press_key();
  write_array(array);
  
  EXIT;
  /*}}}  */
} /* main */

